#include "async.h"
#include "flowsystem.h"
#include <stdio.h>
#include <fcntl.h>
#include <queue>
#include <stdsynthmodule.h>

class StdinRead_impl : public StdinRead_skel, IONotify {
public:
	StdinRead_impl() {
		Dispatcher::the()->ioManager()->watchFD(0, IOType::read, this);
	}
	virtual ~StdinRead_impl() {
		Dispatcher::the()->ioManager()->remove(this, IOType::read);
	}
	void notifyIO(int,int)
	{
		//dprintf("-> StdinRead_impl::process_data\n");
		DataPacket<mcopbyte> *packet = outdata.allocPacket(100);
		char *data = (char *)packet->contents;

		if(fgets(data,100,stdin))
			packet->size = strlen(data);
		else
			packet->size = 0;

		packet->send();
		//dprintf("<- ReadInput::process_data\n");
	}
};

REGISTER_IMPLEMENTATION(StdinRead_impl);

class ToLower_impl : public ToLower_skel {
public:
	void process_indata(DataPacket<mcopbyte> *inpacket)
	{
		DataPacket<mcopbyte> *outpacket = outdata.allocPacket(inpacket->size);

		// convert to lowercase letters
		char *instring = (char *)inpacket->contents;
		char *outstring = (char *)outpacket->contents;

		for(int i=0;i<inpacket->size;i++)
			outstring[i] = tolower(instring[i]);

		inpacket->processed();
		outpacket->send();
	}
};

REGISTER_IMPLEMENTATION(ToLower_impl);

class StdoutWrite_impl : public StdoutWrite_skel {
public:
	void process_indata(DataPacket<mcopbyte> *packet)
	{
		fwrite(packet->contents,sizeof(mcopbyte),packet->size,stdout);
		packet->processed();
	}
	~StdoutWrite_impl() {
		cout << "~StdoutWrite_impl" << endl;
	}
};

REGISTER_IMPLEMENTATION(StdoutWrite_impl);

class ExecuteProcess_impl : public ExecuteProcess_skel, IONotify {
protected:
	FILE *pfile;
	int pfd;
	bool waiting;
	queue< DataPacket<mcopbyte>* > wqueue;
public:
	ExecuteProcess_impl() : pfile(0), waiting(false)
	{
		//
	}

	~ExecuteProcess_impl()
	{
		if(waiting) Dispatcher::the()->ioManager()->remove(this,IOType::read);
	}

	enum { packetCapacity = 4096 };
	void start(const string& commandline)
	{
		assert(!pfile);
		/*
		 * setup new process
		 */
		pfile = popen(commandline.c_str(),"r");
		pfd = fileno(pfile);
		int rc = fcntl(pfd, F_SETFL, O_NONBLOCK);
		assert(rc != -1);
		/*
		 * start streaming
		 */
		outdata.setPull(8, packetCapacity);
	}
	void handle_eof()
	{
		cout << "EOF" << endl;
		/*
		 * cleanup
		 */
		outdata.endPull();
		pclose(pfile);
		pfile = 0;
		pfd = 0;

		/*
		 * remove all packets from the wqueue
		 */
		while(!wqueue.empty())
		{
			DataPacket<mcopbyte> *packet = wqueue.front();
			packet->size = 0;
			packet->send();
			wqueue.pop();
		}
		
		/*
		 * go out of the waiting mode
		 */
		if(waiting)
		{
			Dispatcher::the()->ioManager()->remove(this,IOType::read);
			waiting = false;
		}
	}
	void request_outdata(DataPacket<mcopbyte> *packet)
	{
		if(!waiting)
		{
			packet->size = read(pfd, packet->contents, packetCapacity);
			if(packet->size > 0)
			{
				// got something - done here
				packet->send();
				return;
			}
			Dispatcher::the()->ioManager()->watchFD(pfd,IOType::read,this);
			waiting = true;
		}

		wqueue.push(packet);
	}

	void notifyIO(int,int)
	{
		assert(waiting);

		DataPacket<mcopbyte> *packet = wqueue.front();
		packet->size = read(pfd, packet->contents, packetCapacity);
		assert(packet->size >= 0);
		if(packet->size == 0) {
			handle_eof();
			return;
		}
		packet->send();

		wqueue.pop();

		if(wqueue.empty())
		{
			Dispatcher::the()->ioManager()->remove(this,IOType::read);
			waiting = false;
		}
	}
};

REGISTER_IMPLEMENTATION(ExecuteProcess_impl);

class AsyncTest_impl : public AsyncTest_skel {
protected:
	ExecuteProcess_var execprocess;
	StdinRead_var stdinread;
	ToLower_var tolower;
	StdoutWrite_var stdoutwrite;
	ByteStreamToAudio_var toaudio;
	Object_var play;

	enum { expect_e, expect_n, expect_d, terminated } state;

public:
	AsyncTest_impl() : state(expect_e)
	{
		cout << ">> type end to leave demo <<" << endl;
		
		/*
		execprocess = new ExecuteProcess_impl();
		execprocess->start("testslow");
		stdoutwrite = new StdoutWrite_impl();
		stdoutwrite->_node()->connect("indata",execprocess->_node(),"outdata");
		*/

		//----- create nodes

		// 1. ExecProcess: mpg123 -s
		execprocess = new ExecuteProcess_impl();
	//execprocess->start("mpg123 -s /var/music/mpg/01_beds_are_burning.mp3");
		execprocess->start("mpg123 -s /var/music/mpg/starwars/asgood.mp3");
/*
		stdoutwrite = new StdoutWrite_impl();
		stdoutwrite->_node()->connect("indata",execprocess->_node(),"outdata");
*/

		// 2. A converter of a byte stream to audio data
		toaudio = new ByteStreamToAudio_impl();

		// 3. A play object
		play = ObjectManager::the()->create("Synth_PLAY");
		assert(play);

		//----- connect nodes
		toaudio->_node()->connect("indata",execprocess->_node(),"outdata");
		play->_node()->connect("invalue_left",toaudio->_node(),"left");
		play->_node()->connect("invalue_right",toaudio->_node(),"right");

		//----- start nodes
		toaudio->_node()->start();
		play->_node()->start();

		//----- create a stdin based thing for exiting cleanly
		/*
		stdinread = new StdinRead_impl();
		tolower = new ToLower_impl();
		tolower->_node()->connect("indata",stdinread->_node(),"outdata");
		_node()->connect("indata",tolower->_node(),"outdata");
		*/

		// reference myself
		_copy();
	}
	void process_indata(DataPacket<mcopbyte> *packet)
	{
		for(int i=0;i<packet->size;i++)
			process_char((char)packet->contents[i]);
		packet->processed();

		// dereference myself when terminated
		if(state == terminated) _release();
	}
	void process_char(char c)
	{
		if(state != terminated) {
			if(c == 'e' && state == expect_e) state = expect_n;
			else if(c == 'n' && state == expect_n) state = expect_d;
			else if(c == 'd' && state == expect_d) state = terminated;
			else state = expect_e;
		}
	}
};

REGISTER_IMPLEMENTATION(AsyncTest_impl);
